
#include <GL/glut.h>
#include <vector>
#include <fstream>
#include <iostream>
#include <exception>
#include <algorithm>
#include <string>
#include <cstring>
#include <assert.h>
#include <math.h>
#include <sstream>
#include <windows.h>


#define for if (0); else for

template <class T> 
const T &Min(const T &t1, const T &t2)
{
  return  (t1 < t2) ? t1 : t2;
}

template <class T> 
const T &Max(const T &t1, const T &t2)
{
  return  (t1 < t2) ? t2 : t1;
}

float Randf()
{
  return float(rand())/RAND_MAX;
}

template <class T>
T Sqr(const T &t)
{
  return t*t;
}

template <class T, class A, class U>
std::vector<T, A> &operator <<(std::vector<T, A> &v, const U &obj)
{
  v.push_back(obj);
  return v;
}

const float pi= 3.1415926535897932384626433832795;

//#pragma warning(disable: 4305) //truncation from 'double' to 'float'


class Timer
{
  __int64 start_time;
  double frequency;

  double curr_time;
  double delta_time;
public:
  void Start();
  void Next();
  double GetCurrentTime() { return curr_time; }
  double GetDeltaTime() { return delta_time; }
};


void Timer::Start()
{
  __int64 frequency_int64;
  QueryPerformanceFrequency((LARGE_INTEGER *)&frequency_int64);
  frequency= frequency_int64;
  QueryPerformanceCounter((LARGE_INTEGER *)&start_time);
}

void Timer::Next()
{
  __int64 curr_time_int64;
  QueryPerformanceCounter((LARGE_INTEGER *)&curr_time_int64);
  double new_time= double(curr_time_int64 - start_time)/frequency;
  this->delta_time= new_time - this->curr_time;
  this->curr_time= new_time; 
}

template <class T>
struct DefaultOperations
{
  typedef float Real;
  //there is always guarantee, that (&source != &dest)
  static void Assign(T &dest, const T &source) 
  { 
    dest = source;
  }
  
  static void MulAssign(T &dest, const T &source, Real r) 
  { 
    dest= r*source;
  }

  static void MulAdd(T &dest, const T &source, Real r) 
  {
    dest+= r*source;
  }

  static void SubMulAssign(T &dest, const T &source1, const T &source2, Real r)
  {
    dest= (source2 - source1)*r;
  }

  static void AddMulAssign(T &dest, const T &source1, const T &source2, Real r)
  {
    dest= source1 + r*source2;
  }  

};



template <class T, class Operations= DefaultOperations<T> >
class SplinePlayer;

template <class T, class Operations= DefaultOperations<T> >
class Spline
{
  typedef typename DefaultOperations<T>::Real Real;
  std::vector <T> value;
  std::vector <T> diff;
  std::vector <Real> param;

  friend class SplinePlayer<T, Operations>;
public:
  void Resize(int i);

  void CalculateDiff();
  
  int Size() const { return int(param.size()); }

  Real &ParamAt(int i) 
  { 
    assert( 0 <= i && i < Size());
    return param[i];
  }

  const Real &ParamAt(int i) const
  { 
    assert( 0 <= i && i < Size());
    return param[i];
  }

  T &ValueAt(int i) 
  { 
    assert( 0 <= i && i < Size());
    return value[i];
  }

  const T &ValueAt(int i) const
  { 
    assert( 0 <= i && i < Size());
    return value[i];
  }

  T &DiffAt(int i) 
  {
    assert( 0 <= i && i < Size());
    return diff[i];
  }

  const T &DiffAt(int i) const
  {
    assert( 0 <= i && i < Size());
    return diff[i];
  }
};


template <class T, class Operations>
void Spline<T, Operations>::CalculateDiff()
{
  assert(param.size() >= 2);

  Operations::SubMulAssign(
      DiffAt(0), 
      ValueAt(0), ValueAt(1), 
      1/(ParamAt(1)-ParamAt(0)));

  for (int i=1; i<Size()-1; ++i)
  {
    Operations::SubMulAssign(
      DiffAt(i), 
      ValueAt(i-1), ValueAt(i+1), 
      1/(ParamAt(i+1)-ParamAt(i-1)));
  }

  Operations::SubMulAssign(
    DiffAt(Size()-1), 
    ValueAt(Size()-2), ValueAt(Size()-1), 
    1/(ParamAt(Size()-1)-ParamAt(Size()-2)));
}

template <class T, class Operations>
void Spline<T, Operations>::Resize(int i)
{
  value.resize(i);
  diff.resize(i);
  param.resize(i);
}

template <class T, class Operations>
class SplinePlayer
{
  typedef typename DefaultOperations<T>::Real Real;


  /*
    invariant:
    cur_v == 
    a_koeff*cur_alpha^3 + 
    b_koeff*cur_alpha^2 +
    c_koeff*cur_alpha^1 +
    d_koeff*cur_alpha^1, d_koeff == ValueAt(cur_n), c_koeff == DiffAt(cur_n)*l

    cur_t == ValueAt(cur_n) + l * cur_alpha, l == 1/one_div_l
  */

  T cur_v;
  Real cur_t;
  Real cur_alpha;
  int cur_n;
  T a_koeff;
  T b_koeff;
  bool at_end;
  Real one_div_l;
  const Spline<T, Operations> *spline_ref;  

public:
  SplinePlayer(const Spline<T, Operations> &spline_ref);
  void Start();
  void Next(Real dt); //todo: float t;
  bool AtEnd() { return at_end; }
  T &CurVal() { return cur_v; }
};

template <class T, class Operations>
SplinePlayer<T, Operations>::SplinePlayer(const Spline<T, Operations> &spline_ref):
    spline_ref(&spline_ref)
{
  
}

template <class T, class Operations>
void SplinePlayer<T, Operations>::Start()
{
  cur_n= -1; 
  assert(spline_ref->Size() != 0);
  cur_t= spline_ref->ParamAt(0);
  cur_alpha= 1;
  one_div_l= 1; //any value
  at_end= false;
}

template <class T, class Operations>
void SplinePlayer<T, Operations>::Next(Real dt)
{
  if (AtEnd()) 
    return;
  Real next_t = cur_t + dt;
  Real d_alpha;
  if (next_t >= spline_ref->param[cur_n+1])
  {
    do //do { } while (next_param >= param[cur_n+1])
    {
      ++cur_n; 
      if (cur_n == spline_ref->Size()-1)
      {
        at_end= true;
        cur_v= spline_ref->value[cur_n]; //the last value
        cur_t= spline_ref->param[cur_n]; //the last key 
        cur_alpha= 1;
        //TODO: GetCurrentT -> ...
        return;
      }
    } while (next_t >= spline_ref->param[cur_n+1]);

    Real l= spline_ref->param[cur_n+1] - spline_ref->param[cur_n];
    one_div_l= 1 / l;
    Real offset= next_t - spline_ref->param[cur_n];    
    d_alpha= one_div_l*(offset);

    Operations::MulAssign(a_koeff, spline_ref->ValueAt(cur_n), 2.0);
    Operations::MulAdd(a_koeff, spline_ref->DiffAt(cur_n),     1.0*l);
    Operations::MulAdd(a_koeff, spline_ref->ValueAt(cur_n+1), -2.0);
    Operations::MulAdd(a_koeff, spline_ref->DiffAt(cur_n+1),   1.0*l);

    Operations::MulAssign(b_koeff, spline_ref->ValueAt(cur_n),-3.0);
    Operations::MulAdd(b_koeff, spline_ref->DiffAt(cur_n),    -2.0*l);
    Operations::MulAdd(b_koeff, spline_ref->ValueAt(cur_n+1),  3.0);
    Operations::MulAdd(b_koeff, spline_ref->DiffAt(cur_n+1),  -1.0*l);/**/

    
    /*Operations::SubMulAssign(
      a_koeff,
      spline_ref->value[cur_n], spline_ref->value[cur_n+1], 
      1);    

    Operations::AddMulAssign(
      cur_v,
      spline_ref->value[cur_n], a_koeff,
      d_alpha); /**/

    Operations::Assign(cur_v, spline_ref->ValueAt(cur_n));
    cur_alpha= 0;
  }
  else
  {
    d_alpha= dt*one_div_l;
  }

  /*Operations::MulAdd(
    cur_v,
    a_koeff, d_alpha);*/
  //d_alpha=1/2.0;
  //for (int i=0; i<2; ++i)
  //{

  {
    Real next= cur_alpha+d_alpha;
    Real prev= cur_alpha;
    Operations::MulAdd(
      cur_v, spline_ref->DiffAt(cur_n), (next- prev)/one_div_l);
    
    next*= cur_alpha+d_alpha;
    prev*= cur_alpha;

    Operations::MulAdd(
      cur_v, b_koeff, next- prev);

    next*= cur_alpha+d_alpha;
    prev*= cur_alpha;
    Operations::MulAdd(
      cur_v, a_koeff, next- prev);
  }
  cur_alpha+= d_alpha; 
  cur_t= next_t;

  //}


}


struct vec4
{
  float x,y,z,w;
  explicit vec4(
    float x=0.0f, 
    float y=0.0f, 
    float z=0.0f, 
    float w=1.0f):
    x(x), y(y), z(z), w(w) 
  {}
  float *Ptr() { return reinterpret_cast<float *>(this); }
  const float *Ptr() const { return reinterpret_cast<const float *>(this); }

  vec4 &operator += (const vec4 &right)
  { x+= right.x; y+= right.y;  z+= right.z;   w+= right.w; return *this; }

  vec4 &operator -= (const vec4 &right)
  { x-= right.x; y-= right.y;  z-= right.z;   w-= right.w; return *this; }

  vec4 &operator *= (float factor)
  { x*= factor; y*= factor; z*= factor; w*= factor; return *this; }
};

inline vec4 operator +(const vec4 &left, const vec4 &right)
{
  return vec4(
    left.x + right.x,
    left.y + right.y,
    left.z + right.z,
    left.w + right.w);
}


inline vec4 operator -(const vec4 &left, const vec4 &right)
{
  return vec4(
    left.x - right.x,
    left.y - right.y,
    left.z - right.z,
    left.w - right.w);
}


inline vec4 operator *(const vec4 &left, float factor)
{
  return vec4(
    left.x*factor,
    left.y*factor,
    left.z*factor,
    left.w*factor);
}

inline vec4 operator *(float factor, const vec4 &right)
{
  return vec4(
    factor*right.x,
    factor*right.y,
    factor*right.z,
    factor*right.w);
}

inline vec4 operator -(const vec4 &v)
{
  return vec4(
    -v.x,
    -v.y,
    -v.z,
    -v.w);
}


inline vec4 cross(const vec4 &left, const vec4 &right)
{
    return vec4(
      +left.y*right.z - right.y*left.z,
      -left.x*right.z + right.x*left.z,
      +left.x*right.y - right.x*left.y,
      right.w* left.w);
}

inline vec4 normalize(const vec4 &n)
{
  float t= 1/sqrtf(n.x*n.x + n.y*n.y + n.z*n.z);
  return vec4(
    n.x*t,
    n.y*t,
    n.z*t,
    n.w*t);
}

inline float distance(const vec4 &v1, vec4 const &v2)
{
  return sqrt(
    Sqr(v1.x-v2.x)+
    Sqr(v1.y-v2.y)+
    Sqr(v1.z-v2.z)+
    Sqr(v1.w-v2.w));
}
class Grid
{
  int m;
  int n;
  std::vector <vec4> points;
  std::vector <vec4> normals;
  std::vector <vec4> tex_coords;
public:
  std::vector<vec4> &GetPoints() { return points; }
  const std::vector<vec4> &GetPoints() const { return points; }
  std::vector<vec4> &GetNormals() { return normals; }
  const std::vector<vec4> &GetNormals() const { return normals; }
  std::vector<vec4> &GetTexCoords() { return tex_coords; }
  const std::vector<vec4> &GetTexCoords() const { return tex_coords; }
  vec4 &PointAt(int j, int i) { return points[i*n + j]; } 
  vec4 &NormalAt(int j, int i) { return normals[i*n + j]; } 
  vec4 &TexCoordAt(int j, int i) { return tex_coords[i*n + j]; } 

  int Width() const { return n; }
  int Height() const { return m; }
  
  void Resize(int m, int n);
  void CalcNormals();
  void CalcTexCoords();
};

void Grid::Resize(int m, int n)
{
  if (m!= this->m || n!= this->n)
  {
    points.resize(m*n);
    normals.resize(m*n);
    tex_coords.resize(m*n);
    this->n= n;
    this->m= m;
  }
}

void Grid::CalcTexCoords()
{
  for (int i= 0; i< m; ++i)
  {
    for (int j= 0; j< n; ++j)
    {
      TexCoordAt(j,i)= vec4(
        float(j)/(n-1),
        float(i)/(m-1));
    }
  }
}

void Grid::CalcNormals()
{
  
  for (int i= 1; i< m-1; ++i)
  {
    for (int j= 1; j< n-1; ++j)
    {
      NormalAt(j,i)= 
        normalize( 
          cross(
            PointAt(j,i+1) -  PointAt(j,i-1),
            PointAt(j+1,i) -  PointAt(j-1,i)));
    }
  }

  
  for (int i=0; i<m; ++i)
  {
    bool full_row= (i == 0) || (i == m-1);
    int dj= full_row ? 1 : n-1;
    for (int j=0; j<n; j+= dj)
    {
      NormalAt(j,i)= 
        normalize( 
          cross(
            PointAt(j, Min(m-1, i+1)) -  PointAt(j,Max(i-1, 0)),
            PointAt(Min(n-1, j+1), i) -  PointAt(Max(j-1, 0),i)));
    }
  }
 }

void Draw(Grid &grid)
{
  //bool tex_coords= grid.TexCoordsPresent() != 0;
  glBegin(GL_TRIANGLE_STRIP);
  //glBegin(GL_LINE_STRIP);
    for (int i=0; i<grid.Height()-1; ++i)
    {
      for (int j=0; j<grid.Width(); ++j)
      {
        for (int k=0; k<2; ++k)
        {
          glNormal3fv((grid.NormalAt(j, i+k)).Ptr());
          //if (tex_coords)
          glTexCoord2fv(grid.TexCoordAt(j,i+k).Ptr());

          glVertex3fv(grid.PointAt(j, i+k).Ptr());
        }
      }
      //*//degenerate quad:
      if (i!= grid.Height()-1)
      {
        glVertex3fv(grid.PointAt(grid.Width()-1, i+1).Ptr());
        glVertex3fv(grid.PointAt(0, i+1).Ptr());
      }/**/
    }
  glEnd();

  /* //debug normals: 
  glDisable(GL_LIGHTING);
  glColor3f(0.5,1,0.5);

  glBegin(GL_LINES);
    for (int i=0; i<grid.Height(); ++i)
    {
      for (int j=0; j<grid.Width(); ++j)
      {
        //glNormal3fv(grid.NormalAt(j, i).Ptr());
        glVertex3fv(grid.PointAt(j, i).Ptr());
        glVertex3fv((grid.PointAt(j, i)-0.2*grid.NormalAt(j, i)).Ptr());
      }
    }
  glEnd(); 
  glEnable(GL_LIGHTING);/**/
}


struct GridOperations
{
  typedef float Real;
  //there is always guarantee, that (&source != &dest)

  static void Assign(Grid &dest, const Grid &source) 
  { 
    dest = source;
  }
  
  static void MulAssign(Grid &dest, const Grid &source, Real r) 
  { 
    dest.Resize(source.Height(), source.Width());
    vec4 const *p_s_data= &source.GetPoints().front();
    vec4 *p_d_data= &dest.GetPoints().front();
    
    
    int size= source.Width()*source.Height();
    for (int i= 0; i<size; ++i)
    {
      p_d_data[i]= r*p_s_data[i];
    }
  }

  static void MulAdd(Grid &dest, const Grid &source, Real r) 
  {
    dest.Resize(source.Height(), source.Width());
    vec4 const *p_s_data= &source.GetPoints().front();
    vec4 *p_d_data= &dest.GetPoints().front();    
    
    int size= source.Width()*source.Height();
    for (int i= 0; i<size; ++i)
    {
      p_d_data[i]+= r*p_s_data[i];
    }
  }

  static void SubMulAssign(Grid &dest, const Grid &source1, const Grid &source2, Real r)
  {
    //dest= (source2 - source1)*r;
    dest.Resize(source1.Height(), source1.Width());
    vec4 const *p_s_data1= &source1.GetPoints().front();
    vec4 const *p_s_data2= &source2.GetPoints().front();
    vec4 *p_d_data= &dest.GetPoints().front();
    assert(source1.Width() == source2.Width() && source1.Height() == source2.Height() );
    
    int size= source1.Width()*source1.Height();
    for (int i= 0; i<size; ++i)
    {
      p_d_data[i]= (p_s_data2[i] - p_s_data1[i])*r;
    }
  }

  static void AddMulAssign(Grid &dest, const Grid &source1, const Grid &source2, Real r)
  {
    //dest= source1 + r*source2;
    vec4 const *p_s_data1= &source1.GetPoints().front();
    vec4 const *p_s_data2= &source2.GetPoints().front();
    vec4 *p_d_data= &dest.GetPoints().front();
    dest.Resize(source1.Height(), source1.Width());
    
    int size= source1.Width()*source1.Height();
    for (int i= 0; i<size; ++i)
    {
      p_d_data[i]= p_s_data1[i] + r*p_s_data2[i];
    }
  }  

  /*static void MullAdd(Grid &dest, const Grid &source, Real r)
  {
    dest= r*source;
  }*/
};

void OnIdle()
{
  glutPostRedisplay();
}


void CreateGrid(
  Grid &grid,
  const std::vector<vec4> &edge1,
  const std::vector<vec4> &bone,
  const std::vector<vec4> &edge2,
  int m, int n)
{

  assert(edge1.size()== edge2.size() && edge2.size() == bone.size());
  int size= static_cast<int>(edge1.size());

  grid.Resize(m, n);
  Spline<vec4> leafe_edge1;
  Spline<vec4> leafe_edge2;
  Spline<vec4> leafe_bone;

  leafe_edge1.Resize(size);
  leafe_edge2.Resize(size);
  leafe_bone.Resize(size);

  leafe_edge1.ParamAt(0)= 0;
  leafe_edge2.ParamAt(0)= 0;
  leafe_bone.ParamAt(0)= 0;
  float e1t=0, e2t= 0, bt= 0;
  for (int j= 1; j< size; ++j)
  {
    leafe_edge1.ParamAt(j)= 
      (e1t+= distance(edge1[j], edge1[j-1]));
    leafe_edge2.ParamAt(j)=  
      (e2t+= distance(edge2[j], edge2[j-1]));
    leafe_bone.ParamAt(j)= 
      (bt+= distance(bone[j], bone[j-1]));
  }
  for (int j= 0; j< size; ++j)
  {
    leafe_edge1.ValueAt(j)= edge1[j];
    leafe_edge2.ValueAt(j)= edge2[j];
    leafe_bone.ValueAt(j)= bone[j];
  }

  leafe_edge1.CalculateDiff();
  leafe_edge2.CalculateDiff();
  leafe_bone.CalculateDiff();
  
  SplinePlayer<vec4> leafe_bone_player(leafe_bone);
  leafe_bone_player.Start();
  leafe_bone_player.Next(0);

  SplinePlayer<vec4> leafe_edge1_player(leafe_edge1);
  leafe_edge1_player.Start();
  leafe_edge1_player.Next(0);

  SplinePlayer<vec4> leafe_edge2_player(leafe_edge2);
  leafe_edge2_player.Start();
  leafe_edge2_player.Next(0);

  for (int j=0; j<n; ++j)
  {
    vec4 edge1= leafe_edge1_player.CurVal();
    vec4 middle= leafe_bone_player.CurVal();
    vec4 edge2= leafe_edge2_player.CurVal();;

    Spline<vec4> across;
    across.Resize(3);
    across.ParamAt(0)=0;
    across.ParamAt(1)=1;
    across.ParamAt(2)=2;

    across.ValueAt(0)= edge1;
    across.ValueAt(1)= middle;
    across.ValueAt(2)= edge2;
    across.CalculateDiff();

    SplinePlayer<vec4> across_player(across);
    across_player.Start();
    across_player.Next(0);

    for (int i=0; i<m; ++i)
    {
      //grids.ValueAt(2).PointAt(j,i)= across_player.CurVal();
      grid.PointAt(j,i)= across_player.CurVal();
      across_player.Next(across.ParamAt(2)/(m-1)); 
    }

    /*
    leafe_bone_player.Next(float(leafe_bone.Size())/(n-1));
    leafe_edge1_player.Next(float(leafe_edge1.Size())/(n-1));    
    leafe_edge2_player.Next(float(leafe_edge2.Size())/(n-1));    */

    leafe_bone_player.Next(leafe_bone.ParamAt(size-1)/(n-1));
    leafe_edge1_player.Next(leafe_edge1.ParamAt(size-1)/(n-1));
    leafe_edge2_player.Next(leafe_edge2.ParamAt(size-1)/(n-1));
  }
}

class GreenLeaves
{
  Spline<Grid, GridOperations> grids;
  SplinePlayer<Grid, GridOperations> grid_player;
public:
  GreenLeaves(): grid_player(grids) {} 
  void Update(double dt);
  void Draw();
  void Init(  
    const std::vector<vec4> &edge1,
    const std::vector<vec4> &bone,
    const std::vector<vec4> &edge2,
    float height, float width);
};

void GreenLeaves::Init(  
  const std::vector<vec4> &edge1,
  const std::vector<vec4> &bone,
  const std::vector<vec4> &edge2,
  float height, float width)
{
  grids.Resize(3);
  int m=16, n=16;
  for (int i=0; i<grids.Size(); ++i)
  {
    grids.ValueAt(i).Resize(m, n);
    grids.ParamAt(i)= i*5;  
  }
  
  const float angle= pi/4;
  const float radius= height;
  
  for (int i=0; i<m; ++i)
  {
    const float v= float(i)/(m-1);
    for (int j=0; j<n; ++j)
    {
      //...*0.999+0.0005:
      //prevent calculating wrong normal (of degenerated triangles) at poluses
      const float u= float(j)/(n-1)*0.999+0.0005;
      grids.ValueAt(0).PointAt(j,i)=
        vec4( 
          width*(sin(pi*u)*cos((v-0.5)*angle)),
          height*(1-cos(pi*u)),
          width*(sin(pi*u)*sin((v-0.5)*angle)));
    }
  }

  for (int i=0; i<m; ++i)
  {
    const float v= float(i)/(m-1);
    for (int j=0; j<n; ++j)
    {
      //...*0.999+0.0005:
      //prevent calculating wrong normal (of degenerated triangles) at poluses
      const float u= float(j)/(n-1)*0.999+0.0005;
      grids.ValueAt(1).PointAt(j,i)=
        radius*vec4( 
          u*4,
          v*4,
          0);
    }
  }

  for (int i=0; i<m; ++i)
  {
    const float v= float(i)/m;
    for (int j=0; j<n; ++j)
    {
      const float u= float(j)/m;
      grids.ValueAt(1).PointAt(j,i)=
        radius*vec4( 
          u,
          v,
          0);
    }
  }

  CreateGrid(
    grids.ValueAt(2),
    edge1, bone, edge2,
    m,
    n);

  grids.ValueAt(1)= grids.ValueAt(2);
  grids.CalculateDiff();
  grid_player.Start();
  grid_player.Next(0);
}

void GreenLeaves::Update(double dt)
{
  grid_player.Next(dt);
}

void GreenLeaves::Draw()
{
  glPushMatrix();
  const int asize= 20;
  static float angles[asize]=
  {
    Randf()*360, Randf()*360, Randf()*360, Randf()*360, Randf()*360,
    Randf()*360, Randf()*360, Randf()*360, Randf()*360, Randf()*360,
    Randf()*360, Randf()*360, Randf()*360, Randf()*360, Randf()*360,
    Randf()*360, Randf()*360, Randf()*360, Randf()*360, Randf()*360
  };

  Grid &g= grid_player.CurVal(); //grids.ValueAt(1);//
  g.CalcNormals(); 

  for (int i=0; i<asize; ++i)
  {
    //glRotatef(angles[i], 0, 1, 0);
    glPushMatrix();
    glRotatef(1*angles[i]/asize+i*360.0/asize, 0, 1, 0);
    //glRotatef(360.0/asize, 0, 1, 0);
    glRotatef((angles[(i+3)%asize]-180)/15, 0, 0, 1);
    //glRotatef((angles[(i+7)%asize]-180)/20, 0, 0, 1);
    ::Draw(g); 
    glPopMatrix();
  }
  glPopMatrix();
}

//Spline<vec4> spline;
//SplinePlayer<vec4> spliner(spline);

int window_width, window_height;
Timer timer;
int frame_counter;

GreenLeaves green_leaves;
GreenLeaves flower;
Grid sky;
Grid water;
GLuint sky_tex;

void OnSize(int width, int height)
{
  window_width= width;
  window_height= height;
  glViewport(0, 0, width, height);
}

void SetProjectionMatrix()
{
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  float near_plane= 0.01;
  float far_plane= 1000;
  // ? : is there to avoid division by zero
  float ratio= float(window_height)/(window_width+(window_width ? 0 : 1)); 
  float fov= 1.0; // tan(90(o)/2)
  glFrustum(
    -fov*near_plane, +fov*near_plane, 
    -ratio*near_plane*fov, +ratio*near_plane*fov,
    near_plane, far_plane);
  //gluPerspective(60, window_height/window_width, near_plane, far_plane);
}


#define DBG_GL_PRINT std::cout <<glGetError() <<' '<<__LINE__ <<' ' <<std::endl;
//#undef DBG_GL_PRINT 

void OnDisplay()
{
  ++frame_counter;
  timer.Next();
  //spliner.Next(timer.GetDeltaTime());
  //grid_player.Next(timer.GetDeltaTime());

  if (timer.GetCurrentTime()>10)
    green_leaves.Update(timer.GetDeltaTime());
  if (timer.GetCurrentTime()>25)
    flower.Update(timer.GetDeltaTime());

  glClearColor(0,0,0.4,1);
  
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  SetProjectionMatrix();
  glEnable(GL_LIGHT0);
  glEnable(GL_LIGHTING);
  
  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, 1);

  glEnable(GL_DEPTH_TEST);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  ///glLightfv(GL_LIGHT0, GL_POSITION, vec4(-100,100,-100).Ptr());
  glLightfv(GL_LIGHT0, GL_POSITION, vec4(-10, 10, 100).Ptr());
  
  glTranslatef(0, -1, -5);
  
  glRotatef(sin(timer.GetCurrentTime()/5)*20+40*(1-1/(timer.GetCurrentTime()/10+1)), 1, 0, 0);
  //glRotatef(-90, 1, 0, 0);
  glRotatef(timer.GetCurrentTime()*10-10, 0, 1, 0);
  /*
  glBegin(GL_LINE_STRIP);
    for (int i=0; i<spline.Size(); ++i)
    {
      glVertex3fv(spline.ValueAt(i).Ptr());
    }
  glEnd();/**/

  //SplinePlayer<vec4> sp(spline);

  glPointSize(1);
  //glRotatef(timer.GetCurrentTime()*20, 0, 1, 0);
  glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, vec4(0,1,0.2).Ptr());
  green_leaves.Draw();
  
  glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, vec4(0.7,0.4,0.5).Ptr());
  flower.Draw();

  glPushMatrix();
  float mirror[16]= 
  {
    1, 0,0,0,
    0,-1,0,0,
    0, 0,1,0,
    0, 0,0,1  
  };
  
  glMultMatrixf(mirror);
  glTranslatef(0, 1.0*(1-1/(timer.GetCurrentTime()/2+1)),0);

  glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, (vec4(0,1,0.2)*1.5).Ptr());
  green_leaves.Draw();
  
  glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, (vec4(0.7,0.4,0.5)*1.5).Ptr());
  flower.Draw();

  glPopMatrix();
  {
    glPushMatrix();
    glEnable(GL_TEXTURE_2D);
    glDisable(GL_LIGHTING);
    float rotate[16];
    glGetFloatv(GL_MODELVIEW_MATRIX, rotate);
    rotate[12]= rotate[13]= rotate[14]= 0;
    glLoadMatrixf(rotate);

    glBindTexture(GL_TEXTURE_2D, sky_tex);
    //glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, vec4(0.3, 0.6, 1).Ptr());
    glColor3f(1,1,1); 

    Draw(sky);
    glMultMatrixf(mirror);
    glColor3f(0.5,0.5,0.5); 
    Draw(sky);

    glDisable(GL_TEXTURE_2D);
    glPopMatrix();
  }

  glColor3fv(vec4(0.3, 0.6, 1).Ptr());
  glutSwapBuffers();
}

void OnKey(unsigned char key, int , int )
{
  switch (key)
  {
  case 27: /*esc*/ 
    exit(0); 
    break;
  }  
}



vec4 RandVec() { return vec4(Randf(), Randf(), Randf()); }

void LoadTexture(const char *name)
{
  std::ifstream bmp(name, std::ios::binary);
  BITMAPFILEHEADER bfh;
  BITMAPINFOHEADER bih;
  bmp.read((char*)&bfh, sizeof(bfh));
  bmp.read((char*)&bih, sizeof(bih));

  std::vector<char> data(3*bih.biHeight*bih.biWidth);
  bmp.read(&data[0], 3*bih.biHeight*bih.biWidth);

  gluBuild2DMipmaps(
    GL_TEXTURE_2D, 
    3, bih.biWidth, bih.biHeight, GL_BGR_EXT, GL_UNSIGNED_BYTE, 
    &data[0]);

  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
}

void Init()
{
  timer.Start();
  
  green_leaves.Init(
    std::vector<vec4>() 
      << vec4(0, 0,   -0.01)
      << vec4(1, 0.1, -0.7)
      << vec4(2, 0.2, -0.2)
      << vec4(3, 0.2, -0.01),
    std::vector<vec4>()
      << vec4(0, 0,   0)
      << vec4(1, 0.1, 0)
      << vec4(2, 0.3, 0)
      << vec4(3, 0.2, 0),
    std::vector<vec4>() 
      << vec4(0, 0,   0.01)
      << vec4(1, 0.1, 0.7)
      << vec4(2, 0.2, 0.2)
      << vec4(3, 0.2, 0.01),
    1.5, 1.5);

  /*green_leaves.Init(
    std::vector<vec4>() 
      << vec4(0, 0,   -1)
      << vec4(1, 0.1, -1)
      << vec4(2, 0.2, -1)
      << vec4(3, 0.2, -1),
    std::vector<vec4>()
      << vec4(0, 0,   0)
      << vec4(1, 0.1, 0)
      << vec4(2, 0.3, 0)
      << vec4(3, 0.2, 0),
    std::vector<vec4>() 
      << vec4(0, 0,   1)
      << vec4(1, 0.1, 1)
      << vec4(2, 0.2, 1)
      << vec4(3, 0.2, 1),
    1.5, 0.5);/**/

  flower.Init(
    std::vector<vec4>() 
      << vec4(0,   0.2+0,   -0.11)
      << vec4(1.5/2, 0.2+0.3, -0.3)
      << vec4(2/2, 0.2+0.5, -0.1)
      << vec4(3/2, 0.2+0.6, -0.01),
    std::vector<vec4>()
      << vec4(0,   0.2+0,   0)
      << vec4(1.5/2, 0.2+0.1, 0)
      << vec4(2/2, 0.2+0.6, 0)
      << vec4(3/2, 0.2+0.6, 0),
    std::vector<vec4>() 
      << vec4(0,   0.2+0,   0.11)
      << vec4(1.5/2, 0.2+0.3, 0.3)
      << vec4(2/2, 0.2+0.5, 0.1)
      << vec4(3/2, 0.2+0.6, 0.01),
    1.0,0.3); 

  int sky_sample= 4;
  int sky_size= 40;
  sky.Resize(sky_sample,sky_sample);
  for (int i=0; i<sky_sample; ++i)
  {
    for (int j=0; j<sky_sample; ++j)
    {
      sky.PointAt(j,i).x= (j/float(sky_sample-1)-0.5)*sky_size;
      sky.PointAt(j,i).z= (i/float(sky_sample-1)-0.5)*sky_size;
      if ((j==1 || j==2) && (i==1 || i==2))
        sky.PointAt(j,i).y= 0.5*sky_size;
      else
        sky.PointAt(j,i).y= 0*sky_size;
    }
  }
  sky.CalcNormals();
  sky.CalcTexCoords();

  glGenTextures(1, &sky_tex);
  glBindTexture(GL_TEXTURE_2D, sky_tex);
  LoadTexture("sky.bmp");

  int water_sample= 4;
  int water_size= sky_size;
  water.Resize(water_sample,water_sample);
  for (int i=0; i<water_sample; ++i)
  {
    for (int j=0; j<water_sample; ++j)
    {
      water.PointAt(j,i).x= (j/float(water_sample-1)-0.5)*water_size;
      water.PointAt(j,i).z= (i/float(water_sample-1)-0.5)*water_size;
      water.PointAt(j,i).y= 0;
    }
  }
  water.CalcNormals();
  water.CalcTexCoords();
}


void OnTimer(int)
{
  static double old_time;
  static int old_frame= -1;
  if (frame_counter > 0)
  {
    double cur_time= timer.GetCurrentTime();
    float fps= (frame_counter-old_frame)/(cur_time - old_time);
    std::ostringstream s;
    s <<"flower in ocean"<<'['<<"fps=" << fps <<']';
    old_frame= frame_counter;
    old_time= cur_time;
    glutSetWindowTitle(s.str().c_str());
  }
  else
  {
    old_time= timer.GetCurrentTime();
    old_frame= frame_counter;
  }
  glutTimerFunc(1000, OnTimer, 0);
}

int main(int argc, char *argv[])
{
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
  
  glutInitWindowSize(800, 600);
  glutCreateWindow("flower in ocean");
  glutFullScreen();

  glutDisplayFunc(OnDisplay);
  glutKeyboardFunc(OnKey);
  glutIdleFunc(OnIdle);
  glutReshapeFunc(OnSize);
  glutTimerFunc(1000, OnTimer, 0);

  try
  {
    Init();
  }
  catch (std::exception &ex)
  {
    std::cerr <<ex.what() <<std::endl;
  }

  glutMainLoop();
  return 0;
}
